/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_led.h>

#include "config.h"

const char *
ssid_str(const struct wl_ssid_t *ssid)
{
        static char ssid_str[WL_SSID_MAX_LENGTH];
        
        if (ssid == NULL) {
                *ssid_str = 0;
                return ssid_str;
        }
        
        owl_os_memcpy(ssid_str, ssid->ssid, ssid->len);
        ssid_str[ssid->len] = 0;
        return ssid_str;
}

static void
blink_cb(struct owltmr_tmo *tmo, void *ctx)
{
        if (config_is_enabled()) {
                owlled_toggle(OWL_BOARD_LED7);
        } else { 
                owlled_off(OWL_BOARD_LED7);
                owltmr_deltmo(tmo);
        }
}


void
config_default(struct config *cfg)
{
        owl_os_strcpy(cfg->ssid.ssid, "hdwireless");
        cfg->ssid.len = owl_os_strlen("hdwireless");
        cfg->key[0] = 0;
        cfg->wep = 0;
        IP4_ADDR(&cfg->ip, 192, 168, 1, 100);
        IP4_ADDR(&cfg->mask, 255, 255, 255, 0);
        IP4_ADDR(&cfg->gw, 192, 168, 1, 1);
        IP4_ADDR(&cfg->dns, 192, 168, 1, 1);
        cfg->dhcp = 1;
        cfg->magic = 0xdeadbeef;
}

int
config_apply(struct config *cfg)
{
        int err;
        char *key = NULL;
        uint8_t mode;
        
        if ((err = wlp_get_mode(&mode))) {
                owl_err("fail to get mode err:%d", err);
                return err;
        }

        if (mode != WL_MODE_STA) {
                if ((err = wlp_set_mode(WL_MODE_STA))) {
                        owl_err("fail to set sta mode err:%d", err);
                        return err;
                }
                
                if ((err = wlp_reset())) {
                        owl_err("fail to reset err:%d", err);
                        return err;
                }
        }

        if (owl_os_strlen(cfg->key))
                key = cfg->key;
        
        if ((err = wlp_linkup(&cfg->ssid, key, cfg->wep))) {
                owl_err("failed to set ssid, err:%d", err);
                return err;
        }

        if (cfg->dhcp) {
                if ((err = wlp_set_dhcp(1))) {
                        owl_err("failed to enable dhcp, err:%d", err);
                        return err;
                }
                
        } else {
                if ((err = wlp_set_ipaddr(&cfg->ip, &cfg->mask,
                                          &cfg->gw, &cfg->dns))) {
                        owl_err("failed to set ipaddr, err:%d", err);
                        return err;
                }
        }

        return 0;
}

int
config_enable(void)
{
        int err;
        struct wl_mac_addr_t hwaddr;
        struct ip_addr ip, netmask;
        struct wl_ssid_t ssid; /* default ssid */

        owl_dbg("enter configuration mode");

        if ((err = wlp_get_hwaddr(&hwaddr))) {
                owl_err("fail to get hwaddr err:%d", err);
                return err;
        }

        if ((err = wlp_set_mode(WL_MODE_AP))) {
                owl_err("fail to set ap mode err:%d", err);
                return err;
        }

        if ((err = wlp_reset())) {
                owl_err("fail to reset err:%d", err);
                return err;
        }

        IP4_ADDR(&ip, 192, 168, 1, 1);
        IP4_ADDR(&netmask, 255, 255, 255, 0);

        if ((ssid.len = owl_os_snprintf(ssid.ssid, sizeof(ssid.ssid),
                                        "hdconfig-%02x-%02x-%02x",
                                        hwaddr.octet[3],
                                        hwaddr.octet[4],
                                        hwaddr.octet[5])) < 0) {

                owl_err("fail to format ssid");
                return OWL_FAILURE;
        }
        
        if ((err = wlp_linkup(&ssid, NULL, 0))) {
                owl_err("linkup failed err:%d", err);
                return err;
        }

        if ((err = wlp_set_ipaddr(&ip, &netmask, NULL, NULL))) {
                owl_err("set_ipaddr failed err:%d", err);
                return err;
        }

        if ((err = wlp_set_dhcpd(1))) {
                owl_err("set_dhcpd failed err:%d", err);
                return err;
        }

        owlled_off(OWL_BOARD_LED0);
        owlled_off(OWL_BOARD_LED1);
        owlled_off(OWL_BOARD_LED2);
        owlled_off(OWL_BOARD_LED3);
        owlled_off(OWL_BOARD_LED4);
        owlled_off(OWL_BOARD_LED5);
        owlled_off(OWL_BOARD_LED6);
        owlled_off(OWL_BOARD_LED7);

        if (owltmr_addtmo(1000, 1, blink_cb, NULL) == NULL)
                owl_err("could not add blink tmo");

        return 0;
}

int
config_is_enabled(void)
{
        uint8_t mode;
        int err;
        if ((err = wlp_get_mode(&mode))) {
                owl_err("get_mode failed err:%d", err);
                return err;
        }

        return mode;
}

int
config_write(struct mtd_dev *dev, const struct config *cfg)
{
        int err;
        int len = sizeof(*cfg);

        if ((err = dev->write(dev, 0, (char *) cfg, len)) != len) {
                owl_err("write failed err:%d len:%d", err, len);
                return OWL_ERR_PATH;
        }

        return 0;
}

int
config_read(struct mtd_dev *dev, struct config *cfg)
{
        int err;
        int len = sizeof(*cfg);

        if ((err = dev->read(dev, 0, (char *) cfg, len)) != len) {
                owl_err("read failed err:%d len:%d", err, len);
                return OWL_ERR_PATH;
        }

        if (cfg->magic != 0xdeadbeef) {
                owl_err("invalid magic");
                return OWL_ERR_PATH;
        }
        
        return 0;
}

static void
config_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct mtd_dev *mtd = owlsh_data(sh);

        if (argc < 2)
                goto usage;

        if (!owl_os_strcasecmp(argv[1], "enable")) {
                res = config_enable();

        } else if (!owl_os_strcasecmp(argv[1], "print")) {
                struct config cfg;
                if ((res = config_read(mtd, &cfg)))
                        goto out;
                
                owlsh_printf(sh, "ssid:%s\n\r", ssid_str(&cfg.ssid));
                owlsh_printf(sh, "key:%s\n\r", cfg.key);
                owlsh_printf(sh, "wep:%d\n\r", cfg.wep);
                owlsh_printf(sh, "dhcp:%d\n\r", cfg.dhcp);
                owlsh_printf(sh, "ip:%s\n\r", inet_ntoa(&cfg.ip));
                owlsh_printf(sh, "mask:%s\n\r", inet_ntoa(&cfg.mask));
                owlsh_printf(sh, "gw:%s\n\r", inet_ntoa(&cfg.gw));
                owlsh_printf(sh, "dns:%s\n\r", inet_ntoa(&cfg.dns));
                
        } else if (!owl_os_strcasecmp(argv[1], "apply")) {
                struct config cfg;
                if ((res = config_read(mtd, &cfg)))
                        goto out;
                if ((res = config_apply(&cfg)))
                        goto out;
        } else {
                goto usage;
        }
                
out:        
        owlsh_printf(sh, "%s %d (%s)\n\r",
                     res >= 0 ? "ok" : "err", res, owl_error(res));
        return;

usage:
        owlsh_printf(sh, "usage: config enable|print|apply\n\r");
}

struct owlsh_cmd cmd_config = {
        "config", config_f, "Enter configuration mode"
};

