/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG

#include <stdio.h>

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/drivers/owl_board.h>

#include <util/mtdfile.h>

#include <wl_os.h>

static int mtdfile_erase(struct mtd_dev *dev, int addr, int len);
static int mtdfile_read(struct mtd_dev *dev, int addr, char *buf, int len);
static int mtdfile_write(struct mtd_dev *dev, int addr,
                        const char *buf, int len);
static void mtdfile_suspend(struct mtd_dev *dev);
static void mtdfile_resume(struct mtd_dev *dev);

struct mtd_dev *
mtdfile_create(void)
{
        struct mtd_dev *dev;
        int err;

        owl_info("enter");
        
        if ((dev = owl_os_zalloc(sizeof(*dev))) == NULL) {
                owl_err("could not alloc dev");
                return NULL;
        }

        dev->read = mtdfile_read;
        dev->write = mtdfile_write;
        dev->erase = mtdfile_erase;
        dev->suspend = mtdfile_suspend;
        dev->resume = mtdfile_resume;
        dev->type = MTD_TYPE_FILE;
        dev->flags = MTD_WRITEABLE;
        dev->size = owl_mbit(8);
        dev->writesize = dev->size;
        dev->erasesize[0] = 1;
        dev->name = "file";

        owl_info("name:%s size:%d writesize:%d erasesize:%d,%d,%d,%d",
                 dev->name, dev->size, dev->writesize, dev->erasesize[0],
                 dev->erasesize[1], dev->erasesize[2], dev->erasesize[3]);
        
        return dev;
}

static int
mtdfile_erase(struct mtd_dev *dev, int addr, int len)
{
       
        owl_dbg("dev:%p addr:0x%x len:0x%x", dev, addr, len);
        return 0;
}


static int
mtdfile_read(struct mtd_dev *dev, int addr, char *buf, int len)
{       
        FILE *f;
        int err;

        owl_dbg("dev:%p addr:0x%x buf:%p len:0x%x", dev, addr, buf, len);
        
        f = fopen("mtdfile.bin", "rb");
        if (f == NULL) {
                owl_err("unable to open file");
                return OWL_ERR_PATH;
        }

        if (fseek(f, addr, SEEK_SET) < 0) {
                owl_err("failed to seek");
                fclose(f);
                return OWL_FAILURE;
        }
        
        if ((err = fread(buf, len, 1, f)) != 1) {
                owl_err("fail to read err:%d len:%d", err, len);
                fclose(f);
                return OWL_FAILURE;
        }
                
        fclose(f);
        return len;
}

static int
mtdfile_write(struct mtd_dev *dev, int addr, const char *buf, int len)
{
        FILE *f;
        int err;
        
        owl_dbg("dev:%p addr:0x%x buf:%p len:0x%x", dev, addr, buf, len);

        if (dev == NULL) {
                owl_err("dev was NULL");
                return OWL_ERR_PARAM;
        }
        
        if (len <= 0)
                return 0;

        if (addr + len > dev->size) {
                owl_err("trying to write blocks that are out of range");
                return OWL_ERR_PARAM;
        }

        f = fopen("mtdfile.bin", "wb");
        if (f == NULL) {
                owl_err("unable to open file");
                return OWL_ERR_PATH;
        }

        if (fseek(f, addr, SEEK_SET) < 0) {
                owl_err("failed to seek");
                fclose(f);
                return OWL_FAILURE;
        }

        /* fread() and fwrite() return the number of items successfully read
           or written (i.e., not the number of characters). */
        if ((err = fwrite(buf, len, 1, f)) < 0) {
                owl_err("fail to write err:%d len:%d", err, len);
                fclose(f);
                return OWL_FAILURE;
        }
        
        fclose(f);
        return len;
}


static void
mtdfile_suspend(struct mtd_dev *dev)
{
        return;
}


static void
mtdfile_resume(struct mtd_dev *dev)
{
        return;
}
