/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include <wl_os.h>

#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>

#ifndef OWL_CONFIG_SHELL_PRINTF_LEN
# define OWL_CONFIG_SHELL_PRINTF_LEN 80
#endif

#ifndef OWL_CONFIG_SHELL_MAXARGS
# define OWL_CONFIG_SHELL_MAXARGS 12
#endif

#ifndef OWL_CONFIG_SHELL_CMDLEN
# define OWL_CONFIG_SHELL_CMDLEN  80
#endif

#define in_range(c, lo, up)  ((uint8_t)c >= lo && (uint8_t)c <= up)
#define isprint(c)           in_range(c, 0x20, 0x7f)

struct owlsh_ctx {
        void *data;                             /* context, passed to cmd_f */
        struct owlsh_cmd *cmdlist;              /* list of commands */
        const char *prompt;                     /* prompt */
        void (*outb)(void *ctx, int c);
        void *ctx;                              /* passed to outb */
    
        char *buf;
        int buflen;
        int curpos;
        int escape;

        int flags;
        
        int argc;
        const char *argv[OWL_CONFIG_SHELL_MAXARGS];

        struct owlsh_cmd *defcmd;
};

static void parse_line(char *line, int *argc, const char **argv);
static struct owlsh_cmd *owlsh_find_cmd(struct owlsh_ctx *priv,
                                        const char *label);

void
owlsh_destroy(struct owlsh_ctx *priv)
{
        if (priv == NULL)
                return;

        if (priv->buf)
                owl_os_free(priv->buf);

        owl_os_free(priv);
}

struct owlsh_ctx *
owlsh_create(const char *prompt, 
             void (*outb)(void *ctx, int c), void *ctx, int flags)
{
        struct owlsh_ctx *priv;
        
        if (outb == NULL)
                return NULL;

        priv = owl_os_alloc(sizeof(struct owlsh_ctx));
        if (priv == NULL)
                return NULL;
        owl_os_memset(priv, 0, sizeof(struct owlsh_ctx));

        priv->prompt = prompt ? prompt : "$";

        priv->buflen = OWL_CONFIG_SHELL_CMDLEN;
        priv->buf = owl_os_alloc(priv->buflen);
        if (priv->buf == NULL) {
                owl_os_free(priv);
                return NULL;
        }
        owl_os_memset(priv->buf, 0, priv->buflen);

        priv->flags = flags;
        priv->outb = outb;
        priv->ctx = ctx;

        if (!(priv->flags & OWLSH_NOPROMPT))
                owlsh_printf(priv, "%s ", priv->prompt);
        return priv;
}

void
owlsh_addcmd(struct owlsh_ctx *priv, struct owlsh_cmd *cmd, void *ctx)
{
        cmd->ctx = ctx;
        if (priv->cmdlist == NULL) {
                priv->cmdlist = cmd;

        } else {
                struct owlsh_cmd *iter = priv->cmdlist;
                while (iter->next)
                        iter = iter->next;

                iter->next = cmd;
        }

        cmd->next = NULL;
}


void
owlsh_defcmd(struct owlsh_ctx *priv, struct owlsh_cmd *cmd, void *ctx)
{
        cmd->ctx = ctx;
        priv->defcmd = cmd;
}

void *
owlsh_data(struct owlsh_ctx *priv)
{
        return priv->data;
}

void
owlsh_key(struct owlsh_ctx *priv, char key)
{
        struct owlsh_cmd *cmd = NULL;

        /* check escapechars */
        if (priv->escape) {
                /* save one char for the null termination */
                if (priv->curpos >= OWL_CONFIG_SHELL_CMDLEN - 1) {
                        priv->escape = 0;
                        return;
                }

                switch (key) {
                case 'r':
                        priv->buf[priv->curpos] = '\r';
                        break;
                        
                case 'n':
                        priv->buf[priv->curpos] = '\n';
                        break;

                case '\\':
                        priv->buf[priv->curpos] = '\\';
                        break;
                }
                
                priv->curpos++;
                priv->escape = 0;
                return;
        }

        switch (key) {
        case '\\':
                if (priv->flags & OWLSH_ESCAPECHARS) {
                        priv->escape = 1;
                        break;
                }
                
                /* fall through */

        case '\b':
                if (priv->curpos == 0)
                        break;

                if (!(priv->flags & OWLSH_NOECHO))
                        owlsh_printf(priv, "\b \b");
                priv->curpos--;
                break;

        case '\r':
        case '\n':
                priv->buf[priv->curpos] = '\0';

                if (!(priv->flags & OWLSH_NOECHO))
                        owlsh_printf(priv, "\n\r");

                parse_line(priv->buf, &priv->argc, priv->argv);
                cmd = owlsh_find_cmd(priv, priv->argv[0]);
                if (cmd != NULL) {
                        priv->data = cmd->ctx;
                        cmd->cmd_f(priv, priv->argc, priv->argv);
                } else if (priv->defcmd) {
                        priv->data = priv->defcmd->ctx;
                        priv->defcmd->cmd_f(priv, priv->argc, priv->argv);
                } else if (owl_os_strlen(priv->argv[0])) {
                        owlsh_printf(priv, "command not found:%s\n\r",
                                     priv->argv[0]);
                }

                priv->curpos = 0;

                if (!(priv->flags & OWLSH_NOPROMPT))
                        owlsh_printf(priv, "%s ", priv->prompt);
                break;

        default:
                /* save one char for the null termination */
                if (priv->curpos >= OWL_CONFIG_SHELL_CMDLEN - 1)
                        break;
                
                /* only store visible chars */
                if (key >= ' ' && key < 0x7f) {
                        if (!(priv->flags & OWLSH_NOECHO))
                                owlsh_printf(priv, "%c", key);
                        priv->buf[priv->curpos] = key;
                        priv->curpos++;
                }

                break;
        }
}

void
owlsh_write(struct owlsh_ctx *priv, char *buf, int len)
{
        int i;
        for (i = 0; i < len; i++)
                priv->outb(priv->ctx, buf[i]);
}

void
owlsh_hexdump(struct owlsh_ctx *priv, const char *title, const void *pos,
              int len)
{
        const char *cpos = pos;
	char line[80], *p;
	int i, j;
        owlsh_printf(priv, "%s - hexdump(len=%d):\n\r", title, len); 
	for (i = 0; i < len; i += 16) {
		p = line;
                p += owl_os_snprintf(p, line + sizeof(line) - p, "%04x  ", i);
		for (j = 0; j < 16; j++) {
                        if (i + j < len)
                                owl_os_snprintf(p, line + sizeof(line) - p,
                                                "%02x ", 
                                                (unsigned char)cpos[i + j]);
			else
				owl_os_strcpy(p, "   ");
			p += 3;
		}
                
                owl_os_strcpy(p, ": ");
                for (j = 0; j < 16; j++) {
                        if (i + j < len) {
                                if (isprint((unsigned char) cpos[i + j]))
                                        *p++ = cpos[i + j];
                                else
                                        *p++ = '.';
                        }
                }
		*p = '\0';
		owlsh_printf(priv, "   %s\n\r", line);
	}
} 


int
owlsh_printf(struct owlsh_ctx *priv, const char *fmt, ...)
{
        char *str = NULL;
        va_list args;
        int len;
        char *iter;

        va_start(args, fmt);

        if ((str = owl_os_alloc(OWL_CONFIG_SHELL_PRINTF_LEN)) == NULL)
                return OWL_ERR_MEM;
        
        if ((len = owl_os_vsnprintf(
                     str, OWL_CONFIG_SHELL_PRINTF_LEN, fmt, args)) < 0) {
                owl_os_free(str);
                return OWL_ERR_MEM;
        }

        iter = str;
        while (*iter)
                priv->outb(priv->ctx, *iter++);

        owl_os_free(str);
        return len;
}


/**
 * Caller responsible for freeing the result string.
 *
 */ 
char *
owlsh_joinargs(int argc, const char *argv[])
{
        int i;
        char *str = owl_os_zalloc(1);
        if (str == NULL)
                return NULL;

        /* join arg list */
        for (i = 0; i < argc; i++) {
                int sep;
                void *p;

                /* one byte extra for the null char and one byte for the
                 * extra space added unless this was the last token
                 */
                sep = i < argc - 1 ? 1 : 0;
                p = owl_os_realloc(
                        str,
                        owl_os_strlen(str) + owl_os_strlen(argv[i]) + 1 + sep);
                if (p == NULL) {
                        owl_os_free(str);
                        return NULL;
                }
                        
                str = p;
                owl_os_strcpy(str + owl_os_strlen(str), argv[i]); /* strcat */
                if (sep)
                        owl_os_strcpy(str + owl_os_strlen(str), " ");
        }

        return str;
}

static struct owlsh_cmd *
owlsh_find_cmd(struct owlsh_ctx *priv, const char *label)
{
        struct owlsh_cmd *cmd = NULL;
        struct owlsh_cmd *iter;
        int ambiguous = 0;
        
        if (owl_os_strlen(label) == 0)
                return NULL;
        
        for (iter = priv->cmdlist; iter; iter = iter->next) {

                /* exact match */
                if (!owl_os_strcmp(iter->label, label))
                        return iter;

                /* partial match */
                if (owl_os_strncmp(iter->label, label, owl_os_strlen(label)))
                        continue;

                /* ambiguous */
                if (cmd)
                        ambiguous = 1;

                cmd = iter;
        }
        
        return ambiguous ? NULL : cmd;
}

/* owl_os_strtok() is internal to owl_shell for now */
static char *owl_os_strtok(char *s, const char *delim)
{
        const char *spanp;
        int c, sc;
        char *tok;
        static char *last;
   
        if (s == NULL && (s = last) == NULL)
                return (NULL);

        /*
         * Skip (span) leading delimiters (s += strspn(s, delim), sort of).
         */
cont:
        c = *s++;
        for (spanp = delim; (sc = *spanp++) != 0;) {
                if (c == sc)
                        goto cont;
        }

        if (c == 0) {			/* no non-delimiter characters */
                last = NULL;
                return (NULL);
        }
        tok = s - 1;

        /*
         * Scan token (scan for delimiters: s += strcspn(s, delim), sort of).
         * Note that delim must have one NUL; we stop if we see that, too.
         */
        for (;;) {
                c = *s++;
                spanp = delim;
                do {
                        if ((sc = *spanp++) == c) {
                                if (c == 0)
                                        s = NULL;
                                else
                                        s[-1] = 0;
                                last = s;
                                return (tok);
                        }
                } while (sc != 0);
        }
        /* NOTREACHED */
}

static void
parse_line(char *line, int *argc, const char **argv)
{
        char *tok;
        *argc = 0;
        argv[*argc] = "";
        for (tok = owl_os_strtok(line, " ");
             tok;
             tok = owl_os_strtok(NULL, " ")) {
                argv[(*argc)++] = tok;
                if (*argc == OWL_CONFIG_SHELL_MAXARGS)
                        return;
        }
}

static void
cmd_help_f(struct owlsh_ctx *priv, int argc, const char *argv[])
{
        struct owlsh_cmd *iter;

        (void) argc; /* ununsed */
        (void) argv; /* ununsed */
        
        for (iter = priv->cmdlist; iter; iter = iter->next)
                owlsh_printf(priv, "%-20s %s\n\r", iter->label,
                             iter->help ? iter->help : "");
}

struct owlsh_cmd cmd_help = {
        "help", cmd_help_f, "print this information", NULL, NULL
};
