/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef MTD_H
#define MTD_H

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>

#include <wl_os.h>


/**
 * Flash device interface. This interface can be used directly b.y e.g.
 * dedicated flash file systems or through the MTD translation layer (mtd_ftl.c)
 * if the file system is not designed for flash devices (e.g. FAT).
 *
 * All kinds of flash memory devices should be able to implement this interface.
 * Dataflash devices such as AT45 should simply implement the required functions
 * while nand devices requires an additional stack for ECC, bad block handling
 * etc.
 *
 */


struct mtd_dev;

struct mtd_dev {
        
	const char *name;
	uint32_t size; /* nr_pages * pagesize */
        uint32_t erasesize[4]; /* blocksize(s) */
	uint32_t writesize; /* pagesize */

#define MTD_TYPE_DATAFLASH 1
#define MTD_TYPE_EEPROM    2
#define MTD_TYPE_FILE      3
        uint8_t type;
        
#define MTD_WRITEABLE (1 << 0)
#define MTD_SUSPENDED (1 << 1) /* device is currently in power save mode */
        uint8_t flags;

        /**
         * Erase all blocks that are included in the range addr to addr + len.
         * Typically, the erase method which supports the larges erasesize will
         * be used. The applicable erase method depends on the alignment of
         * addr and len.
         *
         * @param dev is the device context pointer
         * @param addr is the starting address, must be on a erasesize boundary
         * @param len is the size of the area which should be erased, must be
         *        aligned to an erasesize.
         * 
         */
        int (*erase)(struct mtd_dev *dev, uint32_t addr, uint32_t len);

        
        /**
         * Read data from device. Address or length does not have be on any
         * specific boundary.
         *
         * @param dev is the device context pointer
         * @param addr is the starting address
         * @param buf is a pointer to a buffer which will hold the read data
         * @param len is the number of bytes to read.
         *
         */
        int (*read)(struct mtd_dev *dev, uint32_t addr, char *buf,
                    uint32_t len);

        /**
         * Write data to device. Address or length does not have be on any
         * specific boundary. The complete write must stay within one writesize
         * boundary.
         *
         * @param dev is the device context pointer
         * @param addr is the starting address, must be on writesize boundary
         * @param buf is a pointer to a buffer which holds the data to write
         * @param len is the number of bytes to write, must be less than or
         *        equal to writesize bytes.
         * @return the number of bytes written
         *
         */
        int (*write)(struct mtd_dev *dev, uint32_t addr, const char *buf,
                     uint32_t len);
    
        /**
         * Enable device power down state. Device will stay in power down state
         * until accessed using read(), write() or erase() or if power down
         * state is intentionally aborted using resume().
         *
         * Note that the device will not go back into power down state if
         * resumed through read(), write() or erase() until this function is
         * called again.
         *
         * Calling this function while device is suspended have no effect.
         *
         * @param dev is the device context pointer
         *
         */
        void (*suspend)(struct mtd_dev *dev);

        
        /**
         * Resume from device power down state.
         *
         * @param dev is the device context pointer
         *
         */
        void (*resume)(struct mtd_dev *dev);
        
	void *priv;
};

extern struct owlsh_cmd cmd_mtd;

#endif
